from threading import Thread
import autocord
import time


class loops:
    def __init__(self, client):
        self.client = client.client
        self.autocord = client

        # create dictionary to store channels
        # that are being listened into
        self.listeners = {}

    def CHANNEL_LISTENER(self, channel, function, refresh=1, type=0):
        if type == 0:
            Thread(target=lambda: self.CHANNEL_LISTENER(channel, function, type=1)).start()
            return str(hash(channel+id(channel)))
        else:
            scanner = autocord.utils(self.autocord)
            last_scanned = scanner.FETCH_MESSAGE_HISTORY(channel, limit=1)[0]

            # add to listeners
            self.listeners[str(hash(channel+id(channel)))] = {
                'channel': channel,
                'function': function
            }

            # create variable to check if thread needs to be killed
            end = False

            while end is False:
                channel_messages = scanner.FETCH_MESSAGE_HISTORY(channel, limit=100)

                if channel_messages[0].id != last_scanned.id:
                    # loop over messages until last scanned message
                    for i, message in enumerate(channel_messages):
                        if message.id == last_scanned.id:
                            break

                    # loop over new messages
                    for x in range(i):
                        # run function specified in parameter with message object as argument
                        # catch and raise error if function doesn't take arguments
                        try:
                            function(channel_messages[x])
                        except TypeError:
                            raise autocord.ParameterNotFound(f'Could not find parameter for function {str(function)}')

                    # set last scanned
                    last_scanned = channel_messages[0]

                # refresh to check loop again
                for i in range(refresh):
                    if not str(hash(channel+id(channel))) in self.listeners:
                        end = True
                        break
                    time.sleep(1)

    def AUTO_REPLY(self, channel, indicator, response, include_self=False, type=0):
        # indicator is the message that you automatically reply to
        # if a message contains the indicator, it replies
        if type == 0:
            Thread(target=lambda: self.AUTO_REPLY(channel, indicator, response, include_self, type=1)).start()
            self.listeners[str(hash(str(channel)+indicator+response))] = {
                'channel': channel,
                'indicator': indicator,
                'response': response
            }
            return str(hash(str(channel)+indicator+response))
        else:
            # created nested on_message function
            def on_message(message):
                print(message.content)
                if indicator in message.content.lower():
                    # don't respond to self
                    if message.author_id == self.autocord.id:
                        if include_self is True:
                            self.autocord.SEND_MESSAGE(response, channel)

            # create listener
            id = self.CHANNEL_LISTENER(channel, on_message)

            while True:
                if not str(hash(str(channel)+indicator+response)) in self.listeners:
                    self.END(id)
                    break

    def END(self, id):
        # end listener by killing thread
        if str(id) in self.listeners:
            del self.listeners[str(id)]
        else:
            raise autocord.ListenerNotFoundError('Could not find listener with specified id')



class TokenValidationError(Exception):
    pass
class NotFoundError(Exception):
    pass
class BadRequestError(Exception):
    pass
class TaskNotFoundError(Exception):
    pass
class UnauthorizedError(Exception):
    pass
class ListenerNotFoundError(Exception):
    pass
class ParameterNotFound(Exception):
    pass


from .utils import utils
from .loops import loops
from .type import message
from .client import client

# Exceptions
from .exceptions import TokenValidationError
from .exceptions import NotFoundError
from .exceptions import BadRequestError
from .exceptions import TaskNotFoundError
from .exceptions import UnauthorizedError
from .exceptions import ListenerNotFoundError
from .exceptions import ParameterNotFound


import requests

'''
Types class to process different types within Autocord
Types specified in autocord/types.py

message 
    <autocord.message>
channel
    <autocord.channel>
    <autocord.message.channel>
author
    <autocord.author>
    <autocord.message.author>
embed
    <autocord.embed>
    <autocord.message.embeds>
image
    <autocord.image>
    <autocord.message.attachments>
    <autocord.embeds.image>
reaction
    <autocord.reaction>
    <autocord.message.reactions>
emoji
    <autocord.emoji>
    <autocord.reaction.emoji>
'''

class message:
    def __init__(self, json):
        # create message json ect with attributes
        # check for embed or attachments
        try:
            if not json['attachments']:
                self.attachments = []
            else:
                # set attachment attributes
                self.attachments = [image(attachment) for attachment in json['attachments']]
        except KeyError:
            self.attachments = []

        # removed component support

        try:
            if not json['embeds']:
                self.embeds = []
            else:
                # create x amount of embed objects
                self.embeds = [embed(em) for em in json['embeds']]
        except KeyError:
            self.embeds = []

        self.author = author(json['author'])
        self.channel = channel(json['channel_id'])
        self.content = json['content']
        self.edited_timestamp = json['edited_timestamp']
        self.flags = json['flags']
        self.mention_everyone = json['mention_everyone']
        self.id = json['id']

        try:
            self.interaction = interaction(json['interaction'])
        except KeyError:
            self.interaction = None

        # process mentions
        self.mention_roles = [int(i) for i in json['mention_roles']]

        try:
            self.nonce = json['nonce']
        except KeyError:
            self.nonce = None

        self.pinned = json['pinned']

        # get referenced message
        try:
            request = requests.get(f'https://discord.com/api/v9/channels/{self.channel.id}/messages/{json["referenced_message"]["id"]}')
            self.referenced_message = message(request.json())
        except KeyError:
            self.referenced_message = None

        # process message reactions
        try:
            self.reactions = [reaction(re) for re in json['reactions']]
        except KeyError:
            self.reactions = []

        self.timestamp = json['timestamp']
        self.tts = json['tts']
        self.type = json['type']

        try:
            self.webhook_id = json['webhook_id']
        except KeyError:
            self.webhook_id = None


class channel:
    def __init__(self, channel):
        self.id = channel

class author:
    def __init__(self, author):
        # avatar is just discord image id
        self.avatar = author['avatar']
        self.avatar_decoration = author['avatar_decoration']
        self.descriminator = author['discriminator']
        self.id = author['id']
        self.public_flags = author['public_flags']
        self.username = author['username']

class embed:
    def __init__(self, embed):
        try:
            self.description = embed['description']
        except KeyError:
            self.description = None

        try:
            self.image = image(embed['image'])
        except KeyError:
            self.image = None

        try:
            self.title = embed['title']
        except KeyError:
            self.title = None

        try:
            self.url = embed['url']
        except KeyError:
            self.url = None

        self.color = embed['color']
        self.type = embed['type']


class image:
    def __init__(self, image):
        try:
            self.content_type = image['content_type']
        except KeyError:
            self.content_type = None

        try:
            self.filename = image['filename']
        except KeyError:
            self.filename = None

        try:
            self.id = image['id']
        except KeyError:
            self.id = None

        self.height = image['height']
        self.proxy_url = image['prox_url']
        self.url = image['url']
        self.width = image['width']
        self.size = (image['width'], image['height'])

class reaction:
    def __init__(self, reaction):
        self.count = reaction['count']
        self.emoji = emoji(reaction['emoji'])
        # indicates whether or not you reacted to the reaction
        self.me = reaction['me']

class emoji:
    def __init__(self, emoji):
        # id is none for default emojis
        self.id = emoji['id']
        self.name = emoji['name']

class interaction:
    def __init__(self, interaction):
        self.id = interaction['id']
        self.type = interaction['type']
        self.name = interaction['name']
        self.user = author(interaction['user'])



from threading import Thread
import autocord
import requests
import time

class client:
    def __init__(self, token):
        self.token = token

        # create client session
        self.client = requests.Session()
        self.client.headers = {'authorization': self.token}

        # validate token
        request = self.client.post('https://discord.com/api/v9/channels/1/messages', json={'content': ''})
        if request.status_code == 401:
            raise autocord.TokenValidationError('Unable to validate token')

        # get metadata
        self.FETCH_METADATA()
        self.ongoing_tasks = {}

    def __enter__(self):
        return self

    def __exit__(self):
        pass

    def FETCH_METADATA(self):
        # can be used as a function to just refresh metadata
        request = self.client.get('https://discord.com/api/v9/users/@me').json()

        # declare attributes
        self.bio = request['bio']
        self.discriminator = request['discriminator']
        self.email = request['email']
        self.id = request['id']
        self.phone = request['phone']
        self.username = request['username']
        self.verified = request['verified']

    def SEND_MESSAGE(self, message, channel):
        data = {'content': message}
        request = self.client.post(f'https://discord.com/api/v9/channels/{channel}/messages', json=data)

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def END_TASK(self, id):
        # may take up to 1 second for the task to actually be deleted
        # the function checks to see if the task is deleted every second
        if str(id) in self.ongoing_tasks:
            del self.ongoing_tasks[str(id)]
        else:
            raise autocord.TaskNotFoundError('Could not find task with specified id')

    def CREATE_TASK(self, messages: dict, channel, offset=0, type=0):
        # messages should be a dictionary with the message as the key
        # and the interval as the value
        if type == 0:
            # check if function was initiated by user
            # recursively start the thread
            Thread(target=lambda: self.CREATE_TASK(messages, channel, offset, type=1)).start()
            return str(hash(channel+offset))
        else:
            # recursively started
            self.ongoing_tasks[str(hash(channel+offset))] = {
                'messages': messages,
                'channel': channel,
                'offset': offset
            }
            # kill the thread
            end = False
            time.sleep(offset)
            while True:
                # check to break out of while loop
                if end is False:
                    for message in messages:
                        # check to break out of messages loop
                        if end is False:
                            self.SEND_MESSAGE(message, channel)

                            # keep checking if thread needs to be closed
                            # check every one second
                            for i in range(messages[message]):
                                # check if task id exists within tasks list
                                if str(hash(channel+offset)) in self.ongoing_tasks:
                                    time.sleep(1)
                                else:
                                    end = True
                                    break
                        else:
                            break
                else:
                    break

    def CREATE_GROUP(self, users: list):
        # pass in user id's as strings or ints
        # add client id onto users list
        # need to add self id to create a group chat with only the two of you
        users.append(self.id)
        data = {'recipients': [str(user) for user in users]}
        request = self.client.post('https://discord.com/api/v9/users/@me/channels', json=data)

        if request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def LEAVE_GROUP(self, id, silent=True):
        # leave group with silent as parameter
        request = self.client.delete(f'https://discord.com/api/v9/channels/{id}?silent={str(silent).lower()}')

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def BLOCK_USER(self, id):
        # 2 is delete
        request = self.client.put(f'https://discord.com/api/v9/users/@me/relationships/{id}', json={'type': 2})

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def UNBLOCK_USER(self, id):
        request = self.client.delete(f'https://discord.com/api/v9/users/@me/relationships/{id}')

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def FRIEND_REQUEST(self, id):
        request = self.client.put(f'https://discord.com/api/v9/users/@me/relationships/{id}', json={})

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def CHANGE_STATUS(self, message=None, emoji=None):
        data = {'custom_status': {'text': message, 'emoji_name': emoji}}
        request = self.client.patch('https://discord.com/api/v9/users/@me/settings', json=data)

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def CHANGE_BIO(self, message=None, color=None):
        # accent color in hex
        data = {'bio': message, 'accent_color': color}
        request = self.client.patch('https://discord.com/api/v9/users/%40me/profile', json=data)

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request

    def CHANGE_USERNAME(self, username, password):
        data = {'password': password, 'username': username}
        request = self.client.patch('https://discord.com/api/v9/users/@me', json=data)

        if request.status_code == 404:
            raise autocord.NotFoundError(request.json()['message'])
        elif request.status_code == 400:
            raise autocord.BadRequestError(request.json()['message'])
        elif request.status_code == 429:
            raise autocord.UnauthorizedError(request.json()['message'])
        return request


import requests
import autocord
import math
import time


def query_messages(raw):
    # for turning json into <message object>
    messages = []
    for item in raw:
        # append message object
        messages.append(autocord.message(item))
    return messages

class utils:
    def __init__(self, client):
        # pass in autocord.client as parameter
        self.client = client.client
        self.autocord = client

    def FETCH_MESSAGE_HISTORY(self, channel, limit=100, retry=True):
        messages = []
        # request message history

        if limit >= 100:
            request = self.client.get(f'https://discord.com/api/v9/channels/{channel}/messages?limit=100')
        else:
            # fetch leftover amount
            request = self.client.get(f'https://discord.com/api/v9/channels/{channel}/messages?limit={limit % 100}')

        # turn json into <message object>
        messages.extend(query_messages(request.json()))

        if not limit > 100 or limit == 100:
            # return messages if already finished checking limit
            return messages
        else:
            for i in range(math.floor(limit / 100) - 1):
                last_message = messages[len(messages) - 1].id
                request = self.client.get(f'https://discord.com/api/v9/channels/{channel}/messages?limit=100&before={last_message}')

                try:
                    # rate limit, retry after
                    if request.json()['message'] == 'The resource is being rate limited.':
                        # check if user wants to keep retrying
                        if retry is True:
                            time.sleep(request.json()['retry_after'])
                            request = self.client.get(f'https://discord.com/api/v9/channels/{channel}/messages?limit=100&before={last_message}')
                        else:
                            break

                except TypeError:
                    pass

                # append new data
                messages.extend(query_messages(request.json()))

            # return data
            return messages

